/*
    main.c
    Martijn van der Horst - M.G.v.d.Horst@gmail.com
    30.12.2008
    
    Measure the speed of a soccer ball using an IR gate and a net on a PIC16F77. 
    
    Compile:    sdcc -mpic14 -p16f77 main.c
    Simulate:   gpsim -pp16f877 -s embedded.cod embedded.asm

*/

/* Define processor and include header file. */
#define __16f77
#include"pic/pic16f77.h"

/* Setup chip configuration */
typedef unsigned int config;
config at 0x2007 __CONFIG = _CP_OFF &     // Code protection = off
                            _BODEN_ON &   // Brown-out reset = enabled
                            _PWRTE_OFF &  // Power-up timer enabled = off
                            _WDT_OFF &    // Watchdog timer = disabled
                            _HS_OSC;      // Oscillator = high-speed (20Mhz)

void configure(void) {

  // Turn A/D conversion off
  ADON = 0;
  ADCON1 = 0x07;    

  // Configure PORTA
  TRISA = 0x0; // PORTA is an output port
  PORTA = 0xFF;

  // Configure PORTB
  TRISB = 0xFF; // PORTB is an input port
  NOT_RBPU = 0; // Enable the internal pull-up resistors
  RBIE = 1;		// Enable external interrupts on PORTB
  
  // Configure PORTC
  TRISC = 0x0; // PORTC is an output port
  PORTC = 0x0;

  // Configure PORTD
  TRISD = 0x0; // PORTD is an output port
  PORTD = 0xFF;
  
  // Enable timer 1
  TMR1CS = 0; // Use internal clock (FSOSC /4)
  TMR1IE = 1; // Enable timer interrupt
  TMR1ON = 1; // Enable timer

  // Enable timer 0
  T0CS = 0; // Use internal instruction cycle clock (CLKOUT = FSOSC/4)
  T0IE = 1; // Enable timer interrupt
  PSA = 1; // Don't use the pre-scaler

  // Interupt enable
  PIE1 = 1; // Make sure only the timer 1 interrupt is enable in PIE
  PEIE = 1; // Peripheral interrupt enable (for timer 1)
  GIE = 1;  // Global interrupt enable
}

// Display control
static int display[2] = {0, 0};
static int display_phase = 0;

void update_display(void) {
	PORTC = display[display_phase] & 0xFF;
	PORTD = (display[display_phase] >> 8) & 0xFF;
	display_phase = (display_phase + 1) % 2;
	PORTA = (PORTA & 0xFC) | (display_phase+1);
}

// This array links the LCD display digits (0-3) and leds (a-g), to the ports of the processor
// The third index indicates the display_phase
static const int display_leds[4][7][2] = {
    // Most significant LCD display digit
	{
		// LED a
		{0x0004,0x0000},
		// LED b
		{0x0000,0x0002},
		// LED c
		{0x0000,0x0100},
		// LED d
		{0x0008,0x0000},
		// LED e
		{0x0000,0x0008},
		// LED f
		{0x0000,0x0000},
		// LED g
		{0x0000,0x0004}
	},
    // LCD display digit 2
	{
		// LED a
		{0x0000,0x0800},
		// LED b
		{0x0000,0x0200},
		// LED c
		{0x0000,0x0400},
		// LED d
		{0x0400,0x0000},
		// LED e
		{0x0100,0x0000},
		// LED f
		{0x0800,0x0000},
		// LED g
		{0x0200,0x0000}
	},
    // LCD display digit 3
	{
		// LED a
		{0x0010,0x0000},
		// LED b
		{0x0020,0x0000},
		// LED c
		{0x0040,0x0000},
		// LED d
		{0x0000,0x0040},
		// LED e
		{0x0000,0x0080},
		// LED f
		{0x0000,0x0010},
		// LED g
		{0x0000,0x0020}
	},
    // LCD display digit 4
	{
		// LED a
		{0x0000,0x4000},
		// LED b
		{0x0000,0x1000},
		// LED c
		{0x0000,0x2000},
		// LED d
		{0x2000,0x0000},
		// LED e
		{0x0080,0x0000},
		// LED f
		{0x4000,0x0000},
		// LED g
		{0x1000,0x0000}
	}
};

// This array links the numbers (for displaying) to the LEDs (a-g)
// The least significant bit of the value corresponds to LED a
static const short display_digits[10] = {
    // 0
	0x3F,
	// 1
	0x06,
	// 2
	0x5B,
	// 3
	0x4F,
	// 4
	0x66,
	// 5
	0x6D,
	// 6
	0x7D,
	// 7
	0x07,
	// 8
	0x7F,
	// 9
	0x6F
};

static const short display_letters[26] = {
	// A
	0x00,
	// B
	0x00,
	// C
	0x00,
	// D
	0x00,
	// E
	0x79,
	// F
	0x00,
	// G
	0x6F,
	// H
	0x00,
	// I
	0x00,
	// J
	0x00,
	// K
	0x00,
	// L
	0x00,
	// M
	0x00,
	// N
	0x37,
	// O
	0x5C,
	// P
	0x00,
	// Q
	0x00,
	// R
	0x00,
	// S
	0x00,
	// T
	0x78,
	// U
	0x00,
	// V
	0x00,
	// W
	0x00,
	// X
	0x00,
	// Y
	0x00,
	// Z
	0x00
};

void display_digit(int lcd_digit, int value) {
	short int leds = display_digits[value];
	int i;
	int j = 1;
	for (i = 0; i < 7; i++) {
		if ((leds & j) == 0) {
			// Turn led off
			display[0] = (display[0] & (0xFFFF ^ display_leds[lcd_digit][i][0]));
			display[1] = (display[1] & (0xFFFF ^ display_leds[lcd_digit][i][1]));
		} else {
			// Turn led on
			display[0] = (display[0] | display_leds[lcd_digit][i][0]);
			display[1] = (display[1] | display_leds[lcd_digit][i][1]);
		}
		j = j << 1;
	}
}

void display_letter(int lcd_digit, char letter)
{
	short int leds = display_letters[letter-'A'];
	int i;
	int j = 1;
	for (i = 0; i < 7; i++) {
		if ((leds & j) == 0) {
			// Turn led off
			display[0] = (display[0] & (0xFFFF ^ display_leds[lcd_digit][i][0]));
			display[1] = (display[1] & (0xFFFF ^ display_leds[lcd_digit][i][1]));
		} else {
			// Turn led on
			display[0] = (display[0] | display_leds[lcd_digit][i][0]);
			display[1] = (display[1] | display_leds[lcd_digit][i][1]);
		}
		j = j << 1;
	}
}


void clear_digit(int lcd_digit) {
	int i;
	int j = 1;
	for (i = 0; i < 7; i++) {
		// Turn led off
		display[0] = (display[0] & (0xFFFF ^ display_leds[lcd_digit][i][0]));
		display[1] = (display[1] & (0xFFFF ^ display_leds[lcd_digit][i][1]));
		j = j << 1;
	}
}


void display_number(int number)
{
	int i;
	for(i=3;i>=0;i--)
	{
		if (number == 0) clear_digit(i);
		else display_digit(i,number % 10);
		number = number / 10;
	}	
}

// States:
// Wait for the ball to pass the IR detector
#define STATE_WAIT_FOR_BALL 1
// Wait for the ball to hit the net
#define STATE_WAIT_FOR_NET 2
// Ready
#define STATE_ANIMATE_SCORE 3
// Display the score
#define STATE_DISPLAY_SCORE 4

// Distance in meters
#define DISTANCE 10.0

// Maximum score
#define MAX_SCORE 2999

// State variables:
static int state = STATE_WAIT_FOR_BALL;
static int interrupt_count = 0;
static int nr = 0;
static int button_wait = 0;
static int score = 0;
static int scoreStep = 0;
static int scoreDelay = 0;
static unsigned int timer_lo = 0; // Number of timer 0 interrupts
static unsigned int timer_hi = 0; // Number of timer 0 interrupts

// State updated
void state_updated (void) {
	if (state == STATE_WAIT_FOR_BALL) {
		clear_digit(0);
		clear_digit(1);
		display_letter(2,'G');
		display_letter(3,'O');
	}
	if (state == STATE_WAIT_FOR_NET) {
		clear_digit(0);
		display_letter(1,'N');
		display_letter(2,'E');
		display_letter(3,'T');
	}
}

// ISR
void interrupt_handler (void) interrupt 0 {
    // Port B interrupt (for inputs) 
    if (RBIF == 1) {
    	int b = PORTB;
    	
    	// Show the status of the IR detector
    	RA2 = RB4;
    	
    	// Show the status of the Net
    	RA3 = RB5;
    	
    	// If the IR detector connects to ground, the ball passes it
    	if ((RB4 == 0) && (state == STATE_WAIT_FOR_BALL)) {
    		// Start timer
    		timer_lo = 0;
    		timer_hi = 0;
    		state = STATE_WAIT_FOR_NET;
    		state_updated();
    	}
    	
    	// If the net connects to ground, the ball hits it
    	if ((RB5 == 0) && (state == STATE_WAIT_FOR_NET)) {
    		// Calculate score
    		float time = ((timer_hi * 65536.0) + timer_lo) * 0.0000512;
    		if (time > 0) {
    		  float speed = (DISTANCE / time) * 3.6;
    		  if (speed > MAX_SCORE) score = MAX_SCORE; 
      		  else score = speed;
    		} else score = MAX_SCORE;
       		scoreStep = 0;
       		scoreDelay = 1;
    		state = STATE_ANIMATE_SCORE;
    		state_updated();
    	}
    	
    	// If the reset button connects to ground, reset
    	if (RB6 == 0) {
    		state = STATE_WAIT_FOR_BALL;
    		state_updated();
    	}
    	
        RBIF = 0;
    }

    // Timer 1 interrupt (for display)
    // Operating at approx 76.3Hz
    if (TMR1IF == 1) {
        interrupt_count++;
	   	update_display();	    	
	    
	    if (state == STATE_ANIMATE_SCORE)
	    {
		    if ((interrupt_count % scoreDelay) == 0) {
		        scoreStep++;
		    	display_number(scoreStep);
		        if ((score - scoreStep) < 20) scoreDelay=scoreDelay+2;
			    if (scoreStep == score) {
			    	state = STATE_DISPLAY_SCORE;
			    	state_updated();
			    }
		    }
	    }
	    
        TMR1IF = 0;
    }
    
    // Timer 0 interrupt (for time measurement)
    // Operating at 19531.25 Hz
    if (T0IF == 1) {
    	timer_lo++;
    	if (timer_lo == 0) timer_hi++;
    	T0IF = 0;
    }
	
}


void main(void) {
    int i;
  	int nr = 0;

    configure();

    for (i = 0; i < 4 ; i++) display_digit(i,nr);
    update_display();

    state_updated();

    while(1) {          /* Loop forever */
    }
}
